#pragma once

// concept Transform
// {
//   void operator()(LPVOID pDst, LPCVOID pSrc, SIZE_T cb) const;
// };

template <typename TransformT>
void PFrame::copyFrom(const PFrame& source, CRect sourceRect, CRect targetRect, const TransformT& transform)
{
	if (sourceRect.Size() != targetRect.Size())
	{
		std::ostringstream oss;
		oss << "source and target rectangles are of different sizes\r\n"
			"source (" << sourceRect.Width() << ", " << sourceRect.Height() << ")\r\n"
			"target (" << targetRect.Width() << ", " << targetRect.Height() << ")";
		throw std::logic_error(oss.str().c_str());
	}

	if (sourceRect.left < 0 || sourceRect.right > source.header()->biWidth ||
		sourceRect.top < 0 || sourceRect.bottom > abs(source.header()->biHeight))
	{
		std::ostringstream oss;
		oss << "source rectangle: "
			"(" << sourceRect.left  << ", " << sourceRect.top    << ")-"
			"(" << sourceRect.right << ", " << sourceRect.bottom << ")\r\n"
			"out of bounds: "
			"(0, 0)-"
			"(" << source.header()->biWidth << ", " << source.header()->biHeight << ")";
		throw std::logic_error(oss.str().c_str());
	}

	if (targetRect.left < 0)
	{
		sourceRect.left -= targetRect.left;
		targetRect.left = 0;
	}
	if (targetRect.top < 0)
	{
		sourceRect.top -= targetRect.top;
		targetRect.top = 0;
	}
	if (targetRect.right > header()->biWidth)
	{
		sourceRect.right -= targetRect.right - header()->biWidth;
		targetRect.right = header()->biWidth;
	}
	if (targetRect.bottom > abs(header()->biHeight))
	{
		sourceRect.bottom -= targetRect.bottom - abs(header()->biHeight);
		targetRect.bottom = abs(header()->biHeight);
	}

	SIZE_T lineSize = byteFromX(sourceRect.Width());

	for (int ty = targetRect.top, sy = sourceRect.top; ty < targetRect.bottom; ++ty, ++sy)
	{
		transform(ptr(targetRect.left, ty), source.ptr(sourceRect.left, sy), lineSize);
	}
}

